
   ' =============================
   ' Averillsoft Q-Basic 3D Engine
   ' =============================
   ' Version 1.0: 3D clipping & flat shading. 

   ' Written by Richard Averill 1999.
   ' Send e-mail to: nigelolympus@tinet.ie
   ' Keyboard handling code by Exposed Dreams.

   ' Controls: use the arrow keys to move about. Press
   ' space to jump, and use the < and > keys to strafe.
   ' Press escape to quit.

   ' This was written to test out the 3D clipping
   ' (some people call it 'frustrum clipping' routines
   ' that I wrote. It's currently flat shaded but I'm
   ' working on a new version with Gouraud shading. Please
   ' mention me somewhere if you use this code. If you
   ' like it, please give me an e-mail. Note that I haven't
   ' finished writing the clipping code for ClipElevUpper
   ' and ClipElevLower yet.

   ' ---------
   ' Constants
   ' ---------

   CONST radian = (3.14159265359# / 180)

   ' -------------------------
   ' Define New Variable Types
   ' -------------------------

   TYPE Vertex3d
      x AS LONG      ' Original co-ords.
      y AS LONG
      z AS LONG
      x2 AS LONG     ' Translated, rotated
      y2 AS LONG     ' and clipped co-ords.
      z2 AS LONG
   END TYPE

   TYPE TriangleType1
      v1 AS INTEGER
      v2 AS INTEGER
      v3 AS INTEGER
      colour AS INTEGER
   END TYPE

   TYPE TriangleType2
      x1 AS LONG
      y1 AS LONG
      z1 AS LONG
      x2 AS LONG
      y2 AS LONG
      z2 AS LONG
      x3 AS LONG
      y3 AS LONG
      z3 AS LONG
      colour AS INTEGER
      ignore AS INTEGER
   END TYPE

   ' ------------
   ' Declare SUBs
   ' ------------

   DECLARE SUB DoTriangles ()
   DECLARE SUB DrawFlatTop ()
   DECLARE SUB DrawFlatBottom ()
   DECLARE SUB Draw2LHS ()
   DECLARE SUB Draw2RHS ()
   DECLARE SUB ClipPlanLeft ()
   DECLARE SUB ClipPlanRight ()
   DECLARE SUB ClipElevLower ()
   DECLARE SUB ClipElevUpper ()
   DECLARE SUB DoRotAndTrans ()
   DECLARE SUB GetInput ()
   DECLARE SUB DoDepthSorting ()
   DECLARE SUB MakeThirdList ()
   DECLARE SUB RemoveBackfaces ()
   DECLARE SUB DoPerspective ()
   DECLARE SUB DoClipping ()
   DECLARE SUB MakeSecondList ()
   DECLARE SUB DoDrawing ()
   DECLARE SUB DoFrameCounting ()
   DECLARE SUB DoPhysics ()
   DECLARE SUB DoMapMode ()
   DECLARE SUB Wireframe ()
   DECLARE SUB SetupKeyboard ()

   ' -------------------
   ' Dimension variables
   ' -------------------

   DIM SHARED vertex(0 TO 100) AS Vertex3d
   DIM SHARED triangle1(0 TO 100) AS TriangleType1 ' Original list
   DIM SHARED triangle2(0 TO 100) AS TriangleType2 ' Clipped list
   DIM SHARED triangle3(0 TO 100) AS TriangleType2 ' Backface culled list
   DIM SHARED sintable(0 TO 359) AS INTEGER
   DIM SHARED costable(0 TO 359) AS INTEGER
   DIM SHARED keyb%(127), Zkbh$(2), kbonflag%, olkbseg%, olkboff%

   ' ----------------
   ' Common variables
   ' ----------------

   COMMON SHARED noofverts%, nooftris1%, nooftris2%, nooftris3%
   COMMON SHARED viewx%, viewy%, viewz%, xangle%, yangle%, zangle%
   COMMON SHARED quit%, count%, currenttime&, nexttime&, frames%
   COMMON SHARED ax, ay, bx, by, cx, cy, colour
   COMMON SHARED vertvel%, horizvel%, strvel%

   ' ---------------------
   ' Set 3D structure size
   ' ---------------------

   noofverts% = 19
   nooftris1% = 20

   ' ------------------------
   ' Initial Viewer Variables
   ' ------------------------

   viewx% = 0: viewy% = 50: viewz% = 0
   xangle% = 0: yangle% = 0: zangle% = 0

   ' --------------------------------
   ' Setup SINE and COS lookup tables
   ' --------------------------------

   FOR i% = 0 TO 359
      sintable(i%) = INT(SIN(radian * i%) * 512)
      costable(i%) = INT(COS(radian * i%) * 512)
   NEXT

   ' ----------------
   ' Read vertex data
   ' ----------------

   FOR i% = 0 TO (noofverts% - 1)
      READ vertex(i%).x, vertex(i%).y, vertex(i%).z
   NEXT


   ' ------------------------
   ' Make first triangle list
   ' ------------------------

   FOR i% = 0 TO (nooftris1% - 1)
      READ triangle1(i%).v1, triangle1(i%).v2
      READ triangle1(i%).v3, triangle1(i%).colour
   NEXT

   ' ---------------
   ' Set frame timer
   ' ---------------

   nexttime& = (TIMER + 1)

   ' ----------------------
   ' Setup keyboard routine
   ' ----------------------

   SetupKeyboard

   ' ---------
   ' Main loop
   ' ---------

   DO

      ' -------------
      ' Page flipping
      ' -------------

      ap% = (1 - ap%)
      SCREEN 7, , ap%, (1 - ap%)
      WAIT 986, 8
      WAIT 986, 8, 8
     
      LINE (0, 0)-(319, 99), 1, BF
      LINE (0, 100)-(319, 199), 2, BF

      ' -----------
      ' 3D Routines
      ' -----------

      DoRotAndTrans    ' Rotate and translate vertices around viewpoint.
      MakeSecondList   ' Create the second triangle list.
      DoClipping       ' Clip the triangles to the viewing pyramid.
      DoPerspective    ' Perspective transform vertices.
      RemoveBackfaces  ' Flag triangles which are facing away from viewer.
      MakeThirdList    ' Create the third triangle list.
      DoDepthSorting   ' Depth sort the third triangle list.
      DoDrawing        ' Draw the triangles and header.
      DoFrameCounting  ' Frame counter.
      GetInput         ' Get user input.
      DoPhysics        ' Player physics.

   LOOP UNTIL (quit% = 1)

   ' ----
   ' Quit
   ' ----

   DEF SEG = VARSEG(Zkbh$(1))
   CALL Absolute(BYVAL olkbseg%, BYVAL olkboff%, SADD(Zkbh$(1)))
   DEF SEG

   SCREEN 7, 0, 0
   CLS
   PRINT "Averillsoft 3D Engine"
   PRINT "Version 1.0"
   PRINT "Written by Richard Averill 1999."
   PRINT "Send e-mail to: nigelolympus@tinet.ie"
   END

   ' -----------
   ' Vertex data
   ' -----------
   ' x co-ord, y co-ord, z co-ord.

   DATA -50,   0, 450
   DATA  50,   0, 450
   DATA -50, 100, 450
   DATA  50, 100, 450
   DATA -50,   0, 550
   DATA  50,   0, 550
   DATA -50, 100, 550
   DATA  50, 100, 550
   DATA   0, 150, 500
   DATA 100,   0, 450
   DATA 200,   0, 450
   DATA 100,   0, 550
   DATA 200,   0, 550
   DATA 150, 100, 500
   DATA 100,   0, 300
   DATA 200,   0, 300
   DATA 100,   0, 400
   DATA 200,   0, 400
   DATA 150, 100, 350

   ' -------------
   ' Triangle data
   ' -------------
   ' vertex 1, vertex 2, vertex 3, colour.

   ' Vertices must be listed in
   ' anticlockwise order for
   ' backface detection to work.

   DATA 0,  1,  2,  10
   DATA 1,  3,  2,  10
   DATA 1,  7,  3,  9
   DATA 1,  5,  7,  9
   DATA 6,  5,  4,  5
   DATA 7,  5,  6,  5
   DATA 2,  6,  4,  9
   DATA 2,  4,  0,  9
   DATA 2,  3,  8,  14
   DATA 8,  7,  6,  14
   DATA 3,  7,  8,  4
   DATA 8,  6,  2,  4
   DATA 10, 13, 9,  10
   DATA 9,  13, 11, 9
   DATA 11, 13, 12, 3
   DATA 12, 13, 10, 4
   DATA 15, 18, 14, 5
   DATA 14, 18, 16, 6
   DATA 16, 18, 17, 7
   DATA 17, 18, 15, 8

SUB ClipElevLower

   ' ------------------------------------------  /
   ' Clip triangles to lower plane in elevation <*
   ' ------------------------------------------  \*

   FOR i% = 0 TO (nooftris2% - 1)

      IF (triangle2(i%).ignore = 0) THEN

      IF (triangle2(i%).z1 + triangle2(i%).y1) < 0 THEN
         IF (triangle2(i%).z2 + triangle2(i%).y2) < 0 THEN
            IF (triangle2(i%).z3 + triangle2(i%).y3) < 0 THEN
            
               ' ------------------ \    / 1
               ' All points outside  \  /
               ' ------------------   \/ 2 3
            
               triangle2(i%).ignore = 1

            ELSE

               ' ----------------------- \  3 / 1
               ' 3 inside, 1 & 2 outside  \  /
               ' -----------------------   \/ 2

            END IF
         ELSE
            IF (triangle2(i%).z3 + triangle2(i%).y3) < 0 THEN

               ' ----------------------- \  2 / 1
               ' 2 inside, 1 & 3 outside  \  /
               ' -----------------------   \/ 3

            ELSE

               ' ----------------------- \  2 /
               ' 2 & 3 inside, 1 outside  \3 /
               ' -----------------------   \/  1

            END IF
         END IF
      ELSE
         IF (triangle2(i%).z2 + triangle2(i%).y2) < 0 THEN
            IF (triangle2(i%).z3 + triangle2(i%).y3) < 0 THEN
            
               ' ----------------------- \  1 / 2
               ' 1 inside, 2 & 3 outside  \  /
               ' -----------------------   \/ 3

            ELSE
            
               ' ----------------------- \  1 /
               ' 1 & 3 inside, 2 outside  \3 /
               ' -----------------------   \/  2

            END IF
         ELSE
            IF (triangle2(i%).z3 + triangle2(i%).y3) < 0 THEN
            
               ' -----------------------
               ' 1 & 2 inside, 3 outside
               ' -----------------------

            END IF
         END IF
      END IF

      END IF

   NEXT i%

END SUB

SUB ClipElevUpper

   ' ------------------------------------------  /*
   ' Clip triangles to upper plane in elevation <*
   ' ------------------------------------------  \

   FOR i% = 0 TO (nooftris2% - 1)

      IF (triangle2(i%).ignore = 0) THEN

      IF (triangle2(i%).z1 - triangle2(i%).y1) < 0 THEN
         IF (triangle2(i%).z2 - triangle2(i%).y2) < 0 THEN
            IF (triangle2(i%).z3 - triangle2(i%).y3) < 0 THEN
             
               ' ------------------ \    / 1
               ' All points outside  \  /
               ' ------------------   \/ 2 3
             
               triangle2(i%).ignore = 1

            ELSE

               ' ----------------------- \  3 / 1
               ' 3 inside, 1 & 2 outside  \  /
               ' -----------------------   \/ 2


            END IF
         ELSE
            IF (triangle2(i%).z3 - triangle2(i%).y3) < 0 THEN

               ' ----------------------- \  2 / 1
               ' 2 inside, 1 & 3 outside  \  /
               ' -----------------------   \/ 3


            ELSE

               ' ----------------------- \  2 /
               ' 2 & 3 inside, 1 outside  \3 /
               ' -----------------------   \/  1


            END IF
         END IF
      ELSE
         IF (triangle2(i%).z2 - triangle2(i%).y2) < 0 THEN
            IF (triangle2(i%).z3 - triangle2(i%).y3) < 0 THEN
             
               ' ----------------------- \  1 / 2
               ' 1 inside, 2 & 3 outside  \  /
               ' -----------------------   \/ 3


            ELSE
             
               ' ----------------------- \  1 /
               ' 1 & 3 inside, 2 outside  \3 /
               ' -----------------------   \/  2

            END IF
         ELSE
            IF (triangle2(i%).z3 - triangle2(i%).y3) < 0 THEN
             
               ' -----------------------
               ' 1 & 2 inside, 3 outside
               ' -----------------------

            END IF
         END IF
      END IF

      END IF

   NEXT i%

END SUB

SUB ClipPlanLeft
  
   ' ------------------------------------ *\    /
   ' Clip triangles to left plane in plan  *\  /
   ' ------------------------------------   *\/

   FOR i% = 0 TO (nooftris2% - 1)

      IF (triangle2(i%).x1 + triangle2(i%).z1) < 0 THEN
         IF (triangle2(i%).x2 + triangle2(i%).z2) < 0 THEN
            IF (triangle2(i%).x3 + triangle2(i%).z3) < 0 THEN
              
               ' ------------------  1 \    /
               ' All points outside     \  /
               ' ------------------ 2 3  \/

               triangle2(i%).ignore = 1

            ELSE
              
               ' -----------------------   \ 3  /
               ' 3 inside, 1 & 2 outside 1  \  /
               ' -----------------------   2 \/

               TempX = triangle2(i%).x1
               TempZ = triangle2(i%).z1

               a& = (triangle2(i%).x1 * triangle2(i%).z3)
               b& = (triangle2(i%).x3 * triangle2(i%).z1)
               c& = (triangle2(i%).x3 - triangle2(i%).x1)
               d& = (triangle2(i%).z3 - triangle2(i%).z1)

               triangle2(i%).x1 = (a& - b&) \ (c& + d&)
               triangle2(i%).z1 = -triangle2(i%).x1

               a& = (triangle2(i%).y3 - triangle2(i%).y1)

               IF (ABS(c&) >= ABS(d&)) THEN
                  b& = (triangle2(i%).x1 - TempX)
                  triangle2(i%).y1 = triangle2(i%).y1 + (a& * b& \ c&)
               ELSE
                  b& = (triangle2(i%).z1 - TempZ)
                  triangle2(i%).y1 = triangle2(i%).y1 + (a& * b& \ d&)
               END IF

               TempX = triangle2(i%).x2
               TempZ = triangle2(i%).z2

               a& = (triangle2(i%).x2 * triangle2(i%).z3)
               b& = (triangle2(i%).x3 * triangle2(i%).z2)
               c& = (triangle2(i%).x3 - triangle2(i%).x2)
               d& = (triangle2(i%).z3 - triangle2(i%).z2)
              
               triangle2(i%).x2 = (a& - b&) \ (c& + d&)
               triangle2(i%).z2 = -triangle2(i%).x2

               a& = (triangle2(i%).y3 - triangle2(i%).y2)

               IF (ABS(c&) >= ABS(d&)) THEN
                  b& = (triangle2(i%).x2 - TempX)
                  triangle2(i%).y2 = triangle2(i%).y2 + (a& * b& \ c&)
               ELSE
                  b& = (triangle2(i%).z2 - TempZ)
                  triangle2(i%).y2 = triangle2(i%).y2 + (a& * b& \ d&)
               END IF

            END IF
         ELSE
            IF (triangle2(i%).x3 + triangle2(i%).z3) < 0 THEN
              
               ' -----------------------   \ 2  /
               ' 2 inside, 1 & 3 outside 1  \  /
               ' -----------------------   3 \/

               TempX = triangle2(i%).x1
               TempZ = triangle2(i%).z1

               a& = (triangle2(i%).x1 * triangle2(i%).z2)
               b& = (triangle2(i%).x2 * triangle2(i%).z1)
               c& = (triangle2(i%).x2 - triangle2(i%).x1)
               d& = (triangle2(i%).z2 - triangle2(i%).z1)

               triangle2(i%).x1 = (a& - b&) \ (c& + d&)
               triangle2(i%).z1 = -triangle2(i%).x1

               a& = (triangle2(i%).y2 - triangle2(i%).y1)

               IF (ABS(c&) >= ABS(d&)) THEN
                  b& = (triangle2(i%).x1 - TempX)
                  triangle2(i%).y1 = triangle2(i%).y1 + (a& * b& \ c&)
               ELSE
                  b& = (triangle2(i%).z1 - TempZ)
                  triangle2(i%).y1 = triangle2(i%).y1 + (a& * b& \ d&)
               END IF

               TempX = triangle2(i%).x3
               TempZ = triangle2(i%).z3

               a& = (triangle2(i%).x3 * triangle2(i%).z2)
               b& = (triangle2(i%).x2 * triangle2(i%).z3)
               c& = (triangle2(i%).x2 - triangle2(i%).x3)
               d& = (triangle2(i%).z2 - triangle2(i%).z3)
             
               triangle2(i%).x3 = (a& - b&) \ (c& + d&)
               triangle2(i%).z3 = -triangle2(i%).x3

               a& = (triangle2(i%).y2 - triangle2(i%).y3)

               IF (ABS(c&) >= ABS(d&)) THEN
                  b& = (triangle2(i%).x3 - TempX)
                  triangle2(i%).y3 = triangle2(i%).y3 + (a& * b& \ c&)
               ELSE
                  b& = (triangle2(i%).z3 - TempZ)
                  triangle2(i%).y3 = triangle2(i%).y3 + (a& * b& \ d&)
               END IF

            ELSE
              
               ' -----------------------  \ 2  /
               ' 2 & 3 inside, 1 outside   \ 3/
               ' ----------------------- 1  \/

               a& = (triangle2(i%).x1 * triangle2(i%).z2)
               b& = (triangle2(i%).x2 * triangle2(i%).z1)
               c& = (triangle2(i%).x2 - triangle2(i%).x1)
               d& = (triangle2(i%).z2 - triangle2(i%).z1)
              
               triangle2(nooftris2%).x2 = (a& - b&) \ (c& + d&)
               triangle2(nooftris2%).z2 = -triangle2(nooftris2%).x2

               a& = (triangle2(i%).y2 - triangle2(i%).y1)

               IF (ABS(c&) >= ABS(d&)) THEN
                  b& = (triangle2(nooftris2%).x2 - triangle2(i%).x1)
                  triangle2(nooftris2%).y2 = triangle2(i%).y1 + (a& * b& \ c&)
               ELSE
                  b& = (triangle2(nooftris2%).z2 - triangle2(i%).z1)
                  triangle2(nooftris2%).y2 = triangle2(i%).y1 + (a& * b& \ d&)
               END IF

               oldx% = triangle2(i%).x1
               oldz% = triangle2(i%).z1

               a& = (triangle2(i%).x1 * triangle2(i%).z3)
               b& = (triangle2(i%).x3 * triangle2(i%).z1)
               c& = (triangle2(i%).x3 - triangle2(i%).x1)
               d& = (triangle2(i%).z3 - triangle2(i%).z1)
             
               triangle2(i%).x1 = (a& - b&) \ (c& + d&)
               triangle2(i%).z1 = -triangle2(i%).x1

               a& = (triangle2(i%).y3 - triangle2(i%).y1)

               IF (ABS(c&) >= ABS(d&)) THEN
                  b& = (triangle2(i%).x1 - oldx%)
                  triangle2(i%).y1 = triangle2(i%).y1 + (a& * b& \ c&)
               ELSE
                  b& = (triangle2(i%).z1 - oldz%)
                  triangle2(i%).y1 = triangle2(i%).y1 + (a& * b& \ d&)
               END IF

               triangle2(nooftris2%).x1 = triangle2(i%).x1
               triangle2(nooftris2%).y1 = triangle2(i%).y1
               triangle2(nooftris2%).z1 = triangle2(i%).z1
               
               triangle2(nooftris2%).x3 = triangle2(i%).x2
               triangle2(nooftris2%).y3 = triangle2(i%).y2
               triangle2(nooftris2%).z3 = triangle2(i%).z2
               
               triangle2(nooftris2%).ignore = 0
               triangle2(nooftris2%).colour = triangle2(i%).colour

               nooftris2% = (nooftris2% + 1)

            END IF
         END IF
      ELSE
         IF (triangle2(i%).x2 + triangle2(i%).z2) < 0 THEN
            IF (triangle2(i%).x3 + triangle2(i%).z3) < 0 THEN
              
               ' ----------------------- 2 \ 1  /
               ' 1 inside, 2 & 3 outside    \  /
               ' -----------------------   3 \/

               TempX = triangle2(i%).x2
               TempZ = triangle2(i%).z2

               a& = (triangle2(i%).x2 * triangle2(i%).z1)
               b& = (triangle2(i%).x1 * triangle2(i%).z2)
               c& = (triangle2(i%).x1 - triangle2(i%).x2)
               d& = (triangle2(i%).z1 - triangle2(i%).z2)
            
               triangle2(i%).x2 = (a& - b&) \ (c& + d&)
               triangle2(i%).z2 = -triangle2(i%).x2

               a& = (triangle2(i%).y1 - triangle2(i%).y2)

               IF (ABS(c&) >= ABS(d&)) THEN
                  b& = (triangle2(i%).x2 - TempX)
                  triangle2(i%).y2 = triangle2(i%).y2 + (a& * b& \ c&)
               ELSE
                  b& = (triangle2(i%).z2 - TempZ)
                  triangle2(i%).y2 = triangle2(i%).y2 + (a& * b& \ d&)
               END IF

               TempX = triangle2(i%).x3
               TempZ = triangle2(i%).z3

               a& = (triangle2(i%).x3 * triangle2(i%).z1)
               b& = (triangle2(i%).x1 * triangle2(i%).z3)
               c& = (triangle2(i%).x1 - triangle2(i%).x3)
               d& = (triangle2(i%).z1 - triangle2(i%).z3)
           
               triangle2(i%).x3 = (a& - b&) \ (c& + d&)
               triangle2(i%).z3 = -triangle2(i%).x3

               a& = (triangle2(i%).y1 - triangle2(i%).y3)

               IF (ABS(c&) >= ABS(d&)) THEN
                  b& = (triangle2(i%).x3 - TempX)
                  triangle2(i%).y3 = triangle2(i%).y3 + (a& * b& \ c&)
               ELSE
                  b& = (triangle2(i%).z3 - TempZ)
                  triangle2(i%).y3 = triangle2(i%).y3 + (a& * b& \ d&)
               END IF
              
            ELSE
              
               ' -----------------------  \ 1  /
               ' 1 & 3 inside, 2 outside   \ 3/
               ' ----------------------- 2  \/

               a& = (triangle2(i%).x2 * triangle2(i%).z3)
               b& = (triangle2(i%).x3 * triangle2(i%).z2)
               c& = (triangle2(i%).x3 - triangle2(i%).x2)
               d& = (triangle2(i%).z3 - triangle2(i%).z2)
             
               triangle2(nooftris2%).x2 = (a& - b&) \ (c& + d&)
               triangle2(nooftris2%).z2 = -triangle2(nooftris2%).x2

               a& = (triangle2(i%).y3 - triangle2(i%).y2)

               IF (ABS(c&) >= ABS(d&)) THEN
                  b& = (triangle2(nooftris2%).x2 - triangle2(i%).x2)
                  triangle2(nooftris2%).y2 = triangle2(i%).y2 + (a& * b& \ c&)
               ELSE
                  b& = (triangle2(nooftris2%).z2 - triangle2(i%).z2)
                  triangle2(nooftris2%).y2 = triangle2(i%).y2 + (a& * b& \ d&)
               END IF

               TempX = triangle2(i%).x2
               TempZ = triangle2(i%).z2

               a& = (triangle2(i%).x2 * triangle2(i%).z1)
               b& = (triangle2(i%).x1 * triangle2(i%).z2)
               c& = (triangle2(i%).x1 - triangle2(i%).x2)
               d& = (triangle2(i%).z1 - triangle2(i%).z2)
            
               triangle2(i%).x2 = (a& - b&) \ (c& + d&)
               triangle2(i%).z2 = -triangle2(i%).x2

               a& = (triangle2(i%).y1 - triangle2(i%).y2)

               IF (ABS(c&) >= ABS(d&)) THEN
                  b& = (triangle2(i%).x2 - TempX)
                  triangle2(i%).y2 = triangle2(i%).y2 + (a& * b& \ c&)
               ELSE
                  b& = (triangle2(i%).z2 - TempZ)
                  triangle2(i%).y2 = triangle2(i%).y2 + (a& * b& \ d&)
               END IF

               triangle2(nooftris2%).x1 = triangle2(i%).x2
               triangle2(nooftris2%).y1 = triangle2(i%).y2
               triangle2(nooftris2%).z1 = triangle2(i%).z2
              
               triangle2(nooftris2%).x3 = triangle2(i%).x3
               triangle2(nooftris2%).y3 = triangle2(i%).y3
               triangle2(nooftris2%).z3 = triangle2(i%).z3
              
               triangle2(nooftris2%).ignore = 0
               triangle2(nooftris2%).colour = triangle2(i%).colour

               nooftris2% = (nooftris2% + 1)

            END IF
         ELSE
            IF (triangle2(i%).x3 + triangle2(i%).z3) < 0 THEN
              
               ' -----------------------  \ 1  /
               ' 1 & 2 inside, 3 outside   \ 2/
               ' ----------------------- 3  \/

               a& = (triangle2(i%).x3 * triangle2(i%).z1)
               b& = (triangle2(i%).x1 * triangle2(i%).z3)
               c& = (triangle2(i%).x1 - triangle2(i%).x3)
               d& = (triangle2(i%).z1 - triangle2(i%).z3)
            
               triangle2(nooftris2%).x3 = (a& - b&) \ (c& + d&)
               triangle2(nooftris2%).z3 = -triangle2(nooftris2%).x3

               a& = (triangle2(i%).y1 - triangle2(i%).y3)

               IF (ABS(c&) >= ABS(d&)) THEN
                  b& = (triangle2(nooftris2%).x3 - triangle2(i%).x3)
                  triangle2(nooftris2%).y3 = triangle2(i%).y3 + (a& * b& \ c&)
               ELSE
                  b& = (triangle2(nooftris2%).z3 - triangle2(i%).z3)
                  triangle2(nooftris2%).y3 = triangle2(i%).y3 + (a& * b& \ d&)
               END IF

               TempX = triangle2(i%).x3
               TempZ = triangle2(i%).z3

               a& = (triangle2(i%).x3 * triangle2(i%).z2)
               b& = (triangle2(i%).x2 * triangle2(i%).z3)
               c& = (triangle2(i%).x2 - triangle2(i%).x3)
               d& = (triangle2(i%).z2 - triangle2(i%).z3)
           
               triangle2(i%).x3 = (a& - b&) \ (c& + d&)
               triangle2(i%).z3 = -triangle2(i%).x3

               a& = (triangle2(i%).y2 - triangle2(i%).y3)

               IF (ABS(c&) >= ABS(d&)) THEN
                  b& = (triangle2(i%).x3 - TempX)
                  triangle2(i%).y3 = triangle2(i%).y3 + (a& * b& \ c&)
               ELSE
                  b& = (triangle2(i%).z3 - TempZ)
                  triangle2(i%).y3 = triangle2(i%).y3 + (a& * b& \ d&)
               END IF

               triangle2(nooftris2%).x1 = triangle2(i%).x1
               triangle2(nooftris2%).y1 = triangle2(i%).y1
               triangle2(nooftris2%).z1 = triangle2(i%).z1
            
               triangle2(nooftris2%).x2 = triangle2(i%).x3
               triangle2(nooftris2%).y2 = triangle2(i%).y3
               triangle2(nooftris2%).z2 = triangle2(i%).z3
             
               triangle2(nooftris2%).ignore = 0
               triangle2(nooftris2%).colour = triangle2(i%).colour

               nooftris2% = (nooftris2% + 1)

            END IF
         END IF
      END IF
  
   NEXT i%

END SUB

SUB ClipPlanRight

   ' ------------------------------------- \    /*
   ' Clip triangles to right plane in plan  \  /*
   ' -------------------------------------   \/*

   FOR i% = 0 TO (nooftris2% - 1)

      IF (triangle2(i%).ignore = 0) THEN

      IF (triangle2(i%).z1 - triangle2(i%).x1) < 0 THEN
         IF (triangle2(i%).z2 - triangle2(i%).x2) < 0 THEN
            IF (triangle2(i%).z3 - triangle2(i%).x3) < 0 THEN
              
               ' ------------------ \    / 1
               ' All points outside  \  /
               ' ------------------   \/ 2 3
              
               triangle2(i%).ignore = 1

            ELSE

               ' ----------------------- \  3 / 1
               ' 3 inside, 1 & 2 outside  \  /
               ' -----------------------   \/ 2

               TempX = triangle2(i%).x1
               TempZ = triangle2(i%).z1

               a& = (triangle2(i%).x3 * triangle2(i%).z1)
               b& = (triangle2(i%).x1 * triangle2(i%).z3)
               c& = (triangle2(i%).x3 - triangle2(i%).x1)
               d& = (triangle2(i%).z3 - triangle2(i%).z1)
              
               triangle2(i%).x1 = (a& - b&) \ (c& - d&)
               triangle2(i%).z1 = triangle2(i%).x1

               a& = (triangle2(i%).y3 - triangle2(i%).y1)

               IF (ABS(c&) >= ABS(d&)) THEN
                  b& = (triangle2(i%).x1 - TempX)
                  triangle2(i%).y1 = triangle2(i%).y1 + (a& * b& \ c&)
               ELSE
                  b& = (triangle2(i%).z1 - TempZ)
                  triangle2(i%).y1 = triangle2(i%).y1 + (a& * b& \ d&)
               END IF

               TempX = triangle2(i%).x2
               TempZ = triangle2(i%).z2

               a& = (triangle2(i%).x3 * triangle2(i%).z2)
               b& = (triangle2(i%).x2 * triangle2(i%).z3)
               c& = (triangle2(i%).x3 - triangle2(i%).x2)
               d& = (triangle2(i%).z3 - triangle2(i%).z2)
             
               triangle2(i%).x2 = (a& - b&) \ (c& - d&)
               triangle2(i%).z2 = triangle2(i%).x2
               
               a& = (triangle2(i%).y3 - triangle2(i%).y2)

               IF (ABS(c&) >= ABS(d&)) THEN
                  b& = (triangle2(i%).x2 - TempX)
                  triangle2(i%).y2 = triangle2(i%).y2 + (a& * b& \ c&)
               ELSE
                  b& = (triangle2(i%).z2 - TempZ)
                  triangle2(i%).y2 = triangle2(i%).y2 + (a& * b& \ d&)
               END IF

            END IF
         ELSE
            IF (triangle2(i%).z3 - triangle2(i%).x3) < 0 THEN

               ' ----------------------- \  2 / 1
               ' 2 inside, 1 & 3 outside  \  /
               ' -----------------------   \/ 3

               TempX = triangle2(i%).x1
               TempZ = triangle2(i%).z1
               
               a& = (triangle2(i%).x2 * triangle2(i%).z1)
               b& = (triangle2(i%).x1 * triangle2(i%).z2)
               c& = (triangle2(i%).x2 - triangle2(i%).x1)
               d& = (triangle2(i%).z2 - triangle2(i%).z1)
               
               triangle2(i%).x1 = (a& - b&) \ (c& - d&)
               triangle2(i%).z1 = triangle2(i%).x1
               
               a& = (triangle2(i%).y2 - triangle2(i%).y1)

               IF (ABS(c&) >= ABS(d&)) THEN
                  b& = (triangle2(i%).x1 - TempX)
                  triangle2(i%).y1 = triangle2(i%).y1 + (a& * b& \ c&)
               ELSE
                  b& = (triangle2(i%).z1 - TempZ)
                  triangle2(i%).y1 = triangle2(i%).y1 + (a& * b& \ d&)
               END IF
               
               TempX = triangle2(i%).x3
               TempZ = triangle2(i%).z3
               
               a& = (triangle2(i%).x2 * triangle2(i%).z3)
               b& = (triangle2(i%).x3 * triangle2(i%).z2)
               c& = (triangle2(i%).x2 - triangle2(i%).x3)
               d& = (triangle2(i%).z2 - triangle2(i%).z3)
               
               triangle2(i%).x3 = (a& - b&) \ (c& - d&)
               triangle2(i%).z3 = triangle2(i%).x3
               
               a& = (triangle2(i%).y2 - triangle2(i%).y3)

               IF (ABS(c&) >= ABS(d&)) THEN
                  b& = (triangle2(i%).x3 - TempX)
                  triangle2(i%).y3 = triangle2(i%).y3 + (a& * b& \ c&)
               ELSE
                  b& = (triangle2(i%).z3 - TempZ)
                  triangle2(i%).y3 = triangle2(i%).y3 + (a& * b& \ d&)
               END IF
               
            ELSE

               ' ----------------------- \  2 /
               ' 2 & 3 inside, 1 outside  \3 /
               ' -----------------------   \/  1

               a& = (triangle2(i%).x2 * triangle2(i%).z1)
               b& = (triangle2(i%).x1 * triangle2(i%).z2)
               c& = (triangle2(i%).x2 - triangle2(i%).x1)
               d& = (triangle2(i%).z2 - triangle2(i%).z1)
             
               triangle2(nooftris2%).x2 = (a& - b&) \ (c& - d&)
               triangle2(nooftris2%).z2 = triangle2(nooftris2%).x2

               a& = (triangle2(i%).y2 - triangle2(i%).y1)

               IF (ABS(c&) >= ABS(d&)) THEN
                  b& = (triangle2(nooftris2%).x2 - triangle2(i%).x1)
                  triangle2(nooftris2%).y2 = triangle2(i%).y1 + (a& * b& \ c&)
               ELSE
                  b& = (triangle2(nooftris2%).z2 - triangle2(i%).z1)
                  triangle2(nooftris2%).y2 = triangle2(i%).y1 + (a& * b& \ d&)
               END IF

               TempX = triangle2(i%).x1
               TempZ = triangle2(i%).z1

               a& = (triangle2(i%).x3 * triangle2(i%).z1)
               b& = (triangle2(i%).x1 * triangle2(i%).z3)
               c& = (triangle2(i%).x3 - triangle2(i%).x1)
               d& = (triangle2(i%).z3 - triangle2(i%).z1)
            
               triangle2(i%).x1 = (a& - b&) \ (c& - d&)
               triangle2(i%).z1 = triangle2(i%).x1

               a& = (triangle2(i%).y3 - triangle2(i%).y1)

               IF (ABS(c&) >= ABS(d&)) THEN
                  b& = (triangle2(i%).x1 - TempX)
                  triangle2(i%).y1 = triangle2(i%).y1 + (a& * b& \ c&)
               ELSE
                  b& = (triangle2(i%).z1 - TempZ)
                  triangle2(i%).y1 = triangle2(i%).y1 + (a& * b& \ d&)
               END IF

               triangle2(nooftris2%).x1 = triangle2(i%).x1
               triangle2(nooftris2%).y1 = triangle2(i%).y1
               triangle2(nooftris2%).z1 = triangle2(i%).z1
              
               triangle2(nooftris2%).x3 = triangle2(i%).x2
               triangle2(nooftris2%).y3 = triangle2(i%).y2
               triangle2(nooftris2%).z3 = triangle2(i%).z2
              
               triangle2(nooftris2%).ignore = 0
               triangle2(nooftris2%).colour = triangle2(i%).colour

               nooftris2% = (nooftris2% + 1)
               
            END IF
         END IF
      ELSE
         IF (triangle2(i%).z2 - triangle2(i%).x2) < 0 THEN
            IF (triangle2(i%).z3 - triangle2(i%).x3) < 0 THEN
              
               ' ----------------------- \  1 / 2
               ' 1 inside, 2 & 3 outside  \  /
               ' -----------------------   \/ 3

               TempX = triangle2(i%).x2
               TempZ = triangle2(i%).z2

               a& = (triangle2(i%).x1 * triangle2(i%).z2)
               b& = (triangle2(i%).x2 * triangle2(i%).z1)
               c& = (triangle2(i%).x1 - triangle2(i%).x2)
               d& = (triangle2(i%).z1 - triangle2(i%).z2)
             
               triangle2(i%).x2 = (a& - b&) \ (c& - d&)
               triangle2(i%).z2 = triangle2(i%).x2

               a& = (triangle2(i%).y1 - triangle2(i%).y2)

               IF (ABS(c&) >= ABS(d&)) THEN
                  b& = (triangle2(i%).x2 - TempX)
                  triangle2(i%).y2 = triangle2(i%).y2 + (a& * b& \ c&)
               ELSE
                  b& = (triangle2(i%).z2 - TempZ)
                  triangle2(i%).y2 = triangle2(i%).y2 + (a& * b& \ d&)
               END IF

               TempX = triangle2(i%).x3
               TempZ = triangle2(i%).z3

               a& = (triangle2(i%).x1 * triangle2(i%).z3)
               b& = (triangle2(i%).x3 * triangle2(i%).z1)
               c& = (triangle2(i%).x1 - triangle2(i%).x3)
               d& = (triangle2(i%).z1 - triangle2(i%).z3)
            
               triangle2(i%).x3 = (a& - b&) \ (c& - d&)
               triangle2(i%).z3 = triangle2(i%).x3
              
               a& = (triangle2(i%).y1 - triangle2(i%).y3)

               IF (ABS(c&) >= ABS(d&)) THEN
                  b& = (triangle2(i%).x3 - TempX)
                  triangle2(i%).y3 = triangle2(i%).y3 + (a& * b& \ c&)
               ELSE
                  b& = (triangle2(i%).z3 - TempZ)
                  triangle2(i%).y3 = triangle2(i%).y3 + (a& * b& \ d&)
               END IF

            ELSE
              
               ' ----------------------- \  1 /
               ' 1 & 3 inside, 2 outside  \3 /
               ' -----------------------   \/  2

               a& = (triangle2(i%).x1 * triangle2(i%).z2)
               b& = (triangle2(i%).x2 * triangle2(i%).z1)
               c& = (triangle2(i%).x1 - triangle2(i%).x2)
               d& = (triangle2(i%).z1 - triangle2(i%).z2)
            
               triangle2(nooftris2%).x2 = (a& - b&) \ (c& - d&)
               triangle2(nooftris2%).z2 = triangle2(nooftris2%).x2
               
               a& = (triangle2(i%).y1 - triangle2(i%).y2)

               IF (ABS(c&) >= ABS(d&)) THEN
                  b& = (triangle2(nooftris2%).x2 - triangle2(i%).x2)
                  triangle2(nooftris2%).y2 = triangle2(i%).y2 + (a& * b& \ c&)
               ELSE
                  b& = (triangle2(nooftris2%).z2 - triangle2(i%).z2)
                  triangle2(nooftris2%).y2 = triangle2(i%).y2 + (a& * b& \ d&)
               END IF

               TempX = triangle2(i%).x2
               TempZ = triangle2(i%).z2

               a& = (triangle2(i%).x3 * triangle2(i%).z2)
               b& = (triangle2(i%).x2 * triangle2(i%).z3)
               c& = (triangle2(i%).x3 - triangle2(i%).x2)
               d& = (triangle2(i%).z3 - triangle2(i%).z2)
           
               triangle2(i%).x2 = (a& - b&) \ (c& - d&)
               triangle2(i%).z2 = triangle2(i%).x2

               a& = (triangle2(i%).y3 - triangle2(i%).y2)

               IF (ABS(c&) >= ABS(d&)) THEN
                  b& = (triangle2(i%).x2 - TempX)
                  triangle2(i%).y2 = triangle2(i%).y2 + (a& * b& \ c&)
               ELSE
                  b& = (triangle2(i%).z2 - TempZ)
                  triangle2(i%).y2 = triangle2(i%).y2 + (a& * b& \ d&)
               END IF

               triangle2(nooftris2%).x1 = triangle2(i%).x1
               triangle2(nooftris2%).y1 = triangle2(i%).y1
               triangle2(nooftris2%).z1 = triangle2(i%).z1
             
               triangle2(nooftris2%).x3 = triangle2(i%).x2
               triangle2(nooftris2%).y3 = triangle2(i%).y2
               triangle2(nooftris2%).z3 = triangle2(i%).z2
             
               triangle2(nooftris2%).ignore = 0
               triangle2(nooftris2%).colour = triangle2(i%).colour

               nooftris2% = (nooftris2% + 1)

            END IF
         ELSE
            IF (triangle2(i%).z3 - triangle2(i%).x3) < 0 THEN
              
               ' -----------------------
               ' 1 & 2 inside, 3 outside
               ' -----------------------

               a& = (triangle2(i%).x2 * triangle2(i%).z3)
               b& = (triangle2(i%).x3 * triangle2(i%).z2)
               c& = (triangle2(i%).x2 - triangle2(i%).x3)
               d& = (triangle2(i%).z2 - triangle2(i%).z3)
           
               triangle2(nooftris2%).x2 = (a& - b&) \ (c& - d&)
               triangle2(nooftris2%).z2 = triangle2(nooftris2%).x2

               a& = (triangle2(i%).y2 - triangle2(i%).y3)

               IF (ABS(c&) >= ABS(d&)) THEN
                  b& = (triangle2(nooftris2%).x2 - triangle2(i%).x3)
                  triangle2(nooftris2%).y2 = triangle2(i%).y3 + (a& * b& \ c&)
               ELSE
                  b& = (triangle2(nooftris2%).z2 - triangle2(i%).z3)
                  triangle2(nooftris2%).y2 = triangle2(i%).y3 + (a& * b& \ d&)
               END IF

               TempX = triangle2(i%).x3
               TempZ = triangle2(i%).z3

               a& = (triangle2(i%).x1 * triangle2(i%).z3)
               b& = (triangle2(i%).x3 * triangle2(i%).z1)
               c& = (triangle2(i%).x1 - triangle2(i%).x3)
               d& = (triangle2(i%).z1 - triangle2(i%).z3)
          
               triangle2(i%).x3 = (a& - b&) \ (c& - d&)
               triangle2(i%).z3 = triangle2(i%).x3

               a& = (triangle2(i%).y1 - triangle2(i%).y3)

               IF (ABS(c&) >= ABS(d&)) THEN
                  b& = (triangle2(i%).x3 - TempX)
                  triangle2(i%).y3 = triangle2(i%).y3 + (a& * b& \ c&)
               ELSE
                  b& = (triangle2(i%).z3 - TempZ)
                  triangle2(i%).y3 = triangle2(i%).y3 + (a& * b& \ d&)
               END IF

               triangle2(nooftris2%).x1 = triangle2(i%).x2
               triangle2(nooftris2%).y1 = triangle2(i%).y2
               triangle2(nooftris2%).z1 = triangle2(i%).z2
            
               triangle2(nooftris2%).x3 = triangle2(i%).x3
               triangle2(nooftris2%).y3 = triangle2(i%).y3
               triangle2(nooftris2%).z3 = triangle2(i%).z3
            
               triangle2(nooftris2%).ignore = 0
               triangle2(nooftris2%).colour = triangle2(i%).colour

               nooftris2% = (nooftris2% + 1)

            END IF
         END IF
      END IF
 
      END IF

   NEXT i%

END SUB

SUB DoClipping

   ' -----------------------------------
   ' Clip triangles to pyramid of vision
   ' -----------------------------------

   ClipPlanLeft
   ClipPlanRight
   ClipElevUpper
   ClipElevLower

END SUB

SUB DoDepthSorting

   ' --------------------------------------------
   ' Depth sort triangles using average Z co-ords
   ' --------------------------------------------
    
   FOR i% = 0 TO (nooftris3% - 1)
      FOR j% = (i% + 1) TO (nooftris3% - 1)
         IF (triangle3(i%).z1 < triangle3(j%).z1) THEN
             
            SWAP triangle3(i%), triangle3(j%)
          
         END IF
      NEXT
   NEXT

END SUB

SUB DoDrawing

   ' --------------
   ' Draw triangles
   ' --------------

   DoTriangles

   ' -----------
   ' Print title
   ' -----------

   LOCATE 1, 1: PRINT "Ϳ"
   LOCATE 2, 1: PRINT "Averillsoft"
   LOCATE 3, 1: PRINT " 3D Engine "
   LOCATE 4, 1: PRINT ""

END SUB

SUB DoFrameCounting

   ' --------------
   ' Frame counting
   ' --------------

   count% = (count% + 1)
   currenttime& = TIMER
   
   IF (currenttime& > nexttime&) THEN
      nexttime& = (nexttime& + 1)
      frames% = count%
      count% = 0
   END IF

   LOCATE 2, 15: PRINT "FPS:"; frames%

END SUB

SUB DoMapMode

   ' --------------------------
   ' Setup the map mode co-ords
   ' --------------------------

   FOR i% = 0 TO (nooftris2% - 1)

      IF (triangle2(i%).ignore = 0) THEN
       
         triangle2(i%).x1 = (160 + triangle2(i%).x1 / 5)
         triangle2(i%).y1 = (100 - triangle2(i%).z1 / 5)
        
         triangle2(i%).x2 = (160 + triangle2(i%).x2 / 5)
         triangle2(i%).y2 = (100 - triangle2(i%).z2 / 5)
        
         triangle2(i%).x3 = (160 + triangle2(i%).x3 / 5)
         triangle2(i%).y3 = (100 - triangle2(i%).z3 / 5)

      END IF
  
   NEXT

END SUB

SUB DoPerspective

   ' --------------------------
   ' Perspective transformation
   ' --------------------------

   FOR i% = 0 TO (nooftris2% - 1)

      IF (triangle2(i%).ignore = 0) THEN

         IF (triangle2(i%).z1 <= 0) THEN triangle2(i%).z1 = 1
         IF (triangle2(i%).z2 <= 0) THEN triangle2(i%).z2 = 1
         IF (triangle2(i%).z3 <= 0) THEN triangle2(i%).z3 = 1

         triangle2(i%).x1 = 160 + ((triangle2(i%).x1 * 160&) \ triangle2(i%).z1)
         triangle2(i%).y1 = 100 - ((triangle2(i%).y1 * 100&) \ triangle2(i%).z1)
         
         triangle2(i%).x2 = 160 + ((triangle2(i%).x2 * 160&) \ triangle2(i%).z2)
         triangle2(i%).y2 = 100 - ((triangle2(i%).y2 * 100&) \ triangle2(i%).z2)
         
         triangle2(i%).x3 = 160 + ((triangle2(i%).x3 * 160&) \ triangle2(i%).z3)
         triangle2(i%).y3 = 100 - ((triangle2(i%).y3 * 100&) \ triangle2(i%).z3)
    
      END IF

   NEXT

END SUB

SUB DoPhysics

   ' -----------
   ' Move player
   ' -----------

   viewy% = (viewy% + vertvel%)
   IF (viewy% < 50) THEN viewy% = 50: vertvel% = 0
  
   viewz% = viewz% + (costable(yangle%) * horizvel% \ 512)
   viewx% = viewx% + (sintable(yangle%) * horizvel% \ 512)

   sangle% = (yangle% + 90) MOD 360
   viewz% = viewz% + (costable(sangle%) * strvel% \ 512)
   viewx% = viewx% + (sintable(sangle%) * strvel% \ 512)

   ' -------
   ' Gravity
   ' -------

   vertvel% = (vertvel% - 1)
   IF (vertvel% < -50) THEN vertvel% = -50

   ' --------
   ' Friction
   ' --------

   IF (viewy% = 50) THEN
     
      IF (horizvel% > 0) THEN
         horizvel% = (horizvel% - 1)
      ELSEIF (horizvel% < 0) THEN
         horizvel% = (horizvel% + 1)
      END IF
  
      IF (strvel% > 0) THEN
         strvel% = (strvel% - 1)
      ELSEIF (strvel% < 0) THEN
         strvel% = (strvel% + 1)
      END IF

   END IF

END SUB

SUB DoRotAndTrans

   ' -------------------------------
   ' Vertex rotation and translation
   ' -------------------------------

   FOR i% = 0 TO (noofverts% - 1)

      ' --------------------------
      ' Set up temporary variables
      ' --------------------------

      x = (vertex(i%).x - viewx%)
      y = (vertex(i%).y - viewy%)
      z = (vertex(i%).z - viewz%)
    
      ' ----------------
      ' Rotate on y-axis
      ' ----------------

      TempX = (x * costable(yangle%) - z * sintable(yangle%)) * (2 ^ -9)
      TempZ = (x * sintable(yangle%) + z * costable(yangle%)) * (2 ^ -9)

      ' ----------------
      ' Rotate on X-axis
      ' ----------------

      TempY = (y * costable(xangle%) - TempZ * sintable(xangle%)) * (2 ^ -9)
      TempZ = (y * sintable(xangle%) + TempZ * costable(xangle%)) * (2 ^ -9)

      ' ----------------
      ' Rotate on z-axis
      ' ----------------

      OldTempX = TempX
      TempX = (TempX * costable(zangle%) - TempY * sintable(zangle%)) * (2 ^ -9)
      TempY = (OldTempX * sintable(zangle%) + TempY * costable(zangle%)) * (2 ^ -9)
                       
      ' ---------------------
      ' Copy finished co-ords
      ' ---------------------

      vertex(i%).x2 = TempX
      vertex(i%).y2 = TempY
      vertex(i%).z2 = TempZ

   NEXT

END SUB

SUB DoTriangles

   ' ------------------------------------
   ' Work out what type the triangles are
   ' and draw them using the correct SUB.
   ' ------------------------------------
  
   FOR i% = 0 TO (nooftris3% - 1)

      ax = triangle3(i%).x1
      ay = triangle3(i%).y1
      bx = triangle3(i%).x2
      by = triangle3(i%).y2
      cx = triangle3(i%).x3
      cy = triangle3(i%).y3
      colour = triangle3(i%).colour

      IF (ay > cy) THEN
         SWAP ax, cx
         SWAP ay, cy
      END IF
    
      IF (ay > by) THEN
         SWAP ax, bx
         SWAP ay, by
      END IF

      IF (by > cy) THEN
         SWAP bx, cx
         SWAP by, cy
      END IF

      IF (ay = by) AND (by <> cy) THEN
         IF (ax > bx) THEN SWAP ax, bx
         DrawFlatTop
      ELSEIF (by = cy) AND (ay <> cy) THEN
         IF (bx > cx) THEN SWAP bx, cx
         DrawFlatBottom
      ELSEIF (ay <> by) AND (ay <> cy) AND (by <> cy) THEN
         IF (bx < ax + (((cx - ax) * (by - ay) \ (cy - ay)))) THEN
            Draw2LHS
         ELSE
            Draw2RHS
         END IF
      END IF

   NEXT

END SUB

SUB Draw2LHS

   ' -----------------------------------------
   ' Draws a triangle with two left hand sides
   ' -----------------------------------------

   leftx& = (ax * 256)
   rightx& = leftx&
   changeleftx& = ((bx - ax) * 256 \ (by - ay))
   changerightx& = ((cx - ax) * 256 \ (cy - ay))

   FOR y% = ay TO (by - 1)
      leftx% = (leftx& \ 256)
      rightx% = (rightx& \ 256)
      LINE (leftx%, y%)-(rightx%, y%), colour
      leftx& = (leftx& + changeleftx&)
      rightx& = (rightx& + changerightx&)
   NEXT y%

   changeleftx& = ((cx - bx) * 256 \ (cy - by))

   FOR y% = by TO cy
      leftx% = (leftx& \ 256)
      rightx% = (rightx& \ 256)
      LINE (leftx%, y%)-(rightx%, y%), colour
      leftx& = (leftx& + changeleftx&)
      rightx& = (rightx& + changerightx&)
   NEXT y%

END SUB

SUB Draw2RHS

   ' ------------------------------------------
   ' Draws a triangle with two right hand sides
   ' ------------------------------------------

   leftx& = (ax * 256)
   rightx& = leftx&
   changeleftx& = ((cx - ax) * 256 \ (cy - ay))
   changerightx& = ((bx - ax) * 256 \ (by - ay))

   FOR y% = ay TO (by - 1)
      leftx% = (leftx& \ 256)
      rightx% = (rightx& \ 256)
      LINE (leftx%, y%)-(rightx%, y%), colour
      leftx& = (leftx& + changeleftx&)
      rightx& = (rightx& + changerightx&)
   NEXT y%

   changerightx& = (((cx - bx) * 256) \ (cy - by))

   FOR y% = by TO cy
      leftx% = (leftx& \ 256)
      rightx% = (rightx& \ 256)
      LINE (leftx%, y%)-(rightx%, y%), colour
      leftx& = (leftx& + changeleftx&)
      rightx& = (rightx& + changerightx&)
   NEXT y%

END SUB

SUB DrawFlatBottom

   ' -----------------------------------
   ' Draws a triangle with a flat bottom
   ' -----------------------------------

   leftx& = (ax * 256)
   rightx& = leftx&
   changeleftx& = ((bx - ax) * 256 \ (by - ay))
   changerightx& = ((cx - ax) * 256 \ (cy - ay))

   FOR y% = ay TO by
      leftx% = (leftx& \ 256)
      rightx% = (rightx& \ 256)
      LINE (leftx%, y%)-(rightx%, y%), colour
      leftx& = (leftx& + changeleftx&)
      rightx& = (rightx& + changerightx&)
   NEXT y%

END SUB

SUB DrawFlatTop

   ' --------------------------------
   ' Draws a triangle with a flat top
   ' --------------------------------

   leftx& = (ax * 256)
   rightx& = (bx * 256)
   changeleftx& = ((cx - ax) * 256 \ (cy - ay))
   changerightx& = ((cx - bx) * 256 \ (cy - by))

   FOR y% = ay TO cy
      leftx% = (leftx& \ 256)
      rightx% = (rightx& \ 256)
      LINE (leftx%, y%)-(rightx%, y%), colour
      leftx& = (leftx& + changeleftx&)
      rightx& = (rightx& + changerightx&)
   NEXT y%

END SUB

DEFINT A-Z
SUB Fillchar (segment, offset, value, Bytes)

   asm$ = "UN" + CHR$(6) + "V" + CHR$(8) + "F" + CHR$(12) + "" + "؋^" + CHR$(10) + "" + CHR$(23) + "C]"
   DEF SEG = VARSEG(asm$)
   CALL Absolute(BYVAL segment%, BYVAL offset%, BYVAL value%, BYVAL Bytes%, SADD(asm$))

END SUB

DEFSNG A-Z
SUB GetInput

   ' --------------
   ' Get user input
   ' --------------

      ' ---------
      ' Turn left
      ' ---------

      IF keyb%(75) THEN
         yangle% = (yangle% - 1)
             
         IF (yangle% < 0) THEN
            yangle% = (360 + yangle%)
         END IF
      END IF

      ' ----------
      ' Turn right
      ' ----------

      IF keyb%(77) THEN
         yangle% = (yangle% + 1)
             
         IF (yangle% > 359) THEN
            yangle% = (yangle% - 360)
         END IF
      END IF

      ' -----------
      ' Strafe left
      ' -----------

      IF keyb%(51) THEN
         IF viewy% = 50 THEN strvel% = -10
      END IF

      ' ------------
      ' Strafe right
      ' ------------

      IF keyb%(52) THEN
         IF viewy% = 50 THEN strvel% = 10
      END IF

      ' -------------
      ' Move forwards
      ' -------------

      IF keyb%(72) THEN
         IF viewy% = 50 THEN horizvel% = 10
      END IF

      ' --------------
      ' Move backwards
      ' --------------

      IF keyb%(80) THEN
         IF viewy% = 50 THEN horizvel% = -10
      END IF
     
      ' -------
      ' Look up
      ' -------

      IF keyb%(30) THEN
         xangle% = (xangle% + 5)
         IF (xangle% > 359) THEN xangle% = (xangle% - 360)
      END IF

      ' ---------
      ' Look down
      ' ---------

      IF keyb%(44) THEN
         xangle% = (xangle% - 5)
         IF (xangle% < 0) THEN xangle% = (xangle% + 360)
      END IF

      ' ----
      ' Jump
      ' ----

      IF keyb%(57) THEN
         IF viewy% = 50 THEN vertvel% = 20
      END IF

      ' ----
      ' Quit
      ' ----

      IF keyb%(1) THEN
         quit% = 1
      END IF

END SUB

SUB MakeSecondList

   ' ---------------------------
   ' Create second triangle list
   ' ---------------------------

   nooftris2% = nooftris1%

   FOR i% = 0 TO (nooftris2% - 1)
       
      triangle2(i%).x1 = vertex(triangle1(i%).v1).x2
      triangle2(i%).y1 = vertex(triangle1(i%).v1).y2
      triangle2(i%).z1 = vertex(triangle1(i%).v1).z2
       
      triangle2(i%).x2 = vertex(triangle1(i%).v2).x2
      triangle2(i%).y2 = vertex(triangle1(i%).v2).y2
      triangle2(i%).z2 = vertex(triangle1(i%).v2).z2
       
      triangle2(i%).x3 = vertex(triangle1(i%).v3).x2
      triangle2(i%).y3 = vertex(triangle1(i%).v3).y2
      triangle2(i%).z3 = vertex(triangle1(i%).v3).z2

      triangle2(i%).colour = triangle1(i%).colour
      triangle2(i%).ignore = 0

   NEXT

END SUB

SUB MakeThirdList

   ' ----------------------------
   ' Make the third triangle list
   ' ----------------------------

   nooftris3% = 0

   FOR i% = 0 TO (nooftris2% - 1)
       
      IF (triangle2(i%).ignore = 0) THEN
          
         triangle3(nooftris3%).x1 = triangle2(i%).x1
         triangle3(nooftris3%).y1 = triangle2(i%).y1
          
         triangle3(nooftris3%).x2 = triangle2(i%).x2
         triangle3(nooftris3%).y2 = triangle2(i%).y2
          
         triangle3(nooftris3%).x3 = triangle2(i%).x3
         triangle3(nooftris3%).y3 = triangle2(i%).y3

         depth% = ((triangle2(i%).z1 + triangle2(i%).z2 + triangle2(i%).z3) \ 3)
         triangle3(nooftris3%).z1 = depth%

         triangle3(nooftris3%).colour = triangle2(i%).colour

         nooftris3% = (nooftris3% + 1)

      END IF
      
   NEXT

END SUB

SUB RemoveBackfaces

   ' ----------------
   ' Remove backfaces
   ' ----------------

   FOR i% = 0 TO (nooftris2% - 1)
 
      ' -------------------------------
      ' Check if triangle is a backface
      ' -------------------------------
 
      a = (triangle2(i%).x1 - triangle2(i%).x2)
      b = (triangle2(i%).y1 - triangle2(i%).y3)
      c = (triangle2(i%).y1 - triangle2(i%).y2)
      d = (triangle2(i%).x1 - triangle2(i%).x3)

      normal = (a * b) - (c * d)

      ' -----------------------
      ' If it is then ignore it
      ' -----------------------

      IF (normal > 0) THEN
         triangle2(i%).ignore = 1
      END IF
    
   NEXT

END SUB

SUB SetupKeyboard

   ' ------------------------------------------
   ' Sets up the machine code keyboard routine.
   ' This was written by Exposed Dreams.
   ' ------------------------------------------

   Zkbh$(0) = ""
   Zkbh$(0) = Zkbh$(0) + CHR$(&H55)
   Zkbh$(0) = Zkbh$(0) + CHR$(&H89) + CHR$(&HE5)
   Zkbh$(0) = Zkbh$(0) + CHR$(&HB8) + CHR$(&H9) + CHR$(&H35)
   Zkbh$(0) = Zkbh$(0) + CHR$(&HCD) + CHR$(&H21)
   Zkbh$(0) = Zkbh$(0) + CHR$(&H31) + CHR$(&HC0)
   Zkbh$(0) = Zkbh$(0) + CHR$(&H89) + CHR$(&HD8)
   Zkbh$(0) = Zkbh$(0) + CHR$(&H8B) + CHR$(&H5E) + CHR$(&H6)
   Zkbh$(0) = Zkbh$(0) + CHR$(&H89) + CHR$(&H7)
   Zkbh$(0) = Zkbh$(0) + CHR$(&H8C) + CHR$(&HC0)
   Zkbh$(0) = Zkbh$(0) + CHR$(&H8B) + CHR$(&H5E) + CHR$(&H8)
   Zkbh$(0) = Zkbh$(0) + CHR$(&H89) + CHR$(&H7)
   Zkbh$(0) = Zkbh$(0) + CHR$(&H5D)
   Zkbh$(0) = Zkbh$(0) + CHR$(&HCA) + CHR$(&H4) + CHR$(&H0)

   Zkbh$(1) = ""
   Zkbh$(1) = Zkbh$(1) + CHR$(&H55)
   Zkbh$(1) = Zkbh$(1) + CHR$(&H1E)
   Zkbh$(1) = Zkbh$(1) + CHR$(&H89) + CHR$(&HE5)
   Zkbh$(1) = Zkbh$(1) + CHR$(&H8B) + CHR$(&H46) + CHR$(&HA)
   Zkbh$(1) = Zkbh$(1) + CHR$(&H8E) + CHR$(&HD8)
   Zkbh$(1) = Zkbh$(1) + CHR$(&H8B) + CHR$(&H56) + CHR$(&H8)
   Zkbh$(1) = Zkbh$(1) + CHR$(&HB8) + CHR$(&H9) + CHR$(&H25)
   Zkbh$(1) = Zkbh$(1) + CHR$(&HCD) + CHR$(&H21)
   Zkbh$(1) = Zkbh$(1) + CHR$(&H1F)
   Zkbh$(1) = Zkbh$(1) + CHR$(&H5D)
   Zkbh$(1) = Zkbh$(1) + CHR$(&HCA) + CHR$(&H4) + CHR$(&H0)

   Zkbh$(2) = ""
   Zkbh$(2) = Zkbh$(2) + CHR$(&H50)
   Zkbh$(2) = Zkbh$(2) + CHR$(&H53)
   Zkbh$(2) = Zkbh$(2) + CHR$(&H1E)
   Zkbh$(2) = Zkbh$(2) + CHR$(&HFB)
   Zkbh$(2) = Zkbh$(2) + CHR$(&HB8) + MKI$(VARSEG(keyb%(0)))
   Zkbh$(2) = Zkbh$(2) + CHR$(&H31) + CHR$(&HDB)
   Zkbh$(2) = Zkbh$(2) + CHR$(&H8E) + CHR$(&HD8)
   Zkbh$(2) = Zkbh$(2) + CHR$(&HE4) + CHR$(&H60)
   Zkbh$(2) = Zkbh$(2) + CHR$(&HD0) + CHR$(&HD0)
   Zkbh$(2) = Zkbh$(2) + CHR$(&H88) + CHR$(&HC3)
   Zkbh$(2) = Zkbh$(2) + CHR$(&HF) + CHR$(&H93) + CHR$(&HC0)
   Zkbh$(2) = Zkbh$(2) + CHR$(&H81) + CHR$(&HC3) + MKI$(VARPTR(keyb%(0)))
   Zkbh$(2) = Zkbh$(2) + CHR$(&H88) + CHR$(&H7)
   Zkbh$(2) = Zkbh$(2) + CHR$(&HE4) + CHR$(&H61)
   Zkbh$(2) = Zkbh$(2) + CHR$(&H80) + CHR$(&HCC) + CHR$(&H82)
   Zkbh$(2) = Zkbh$(2) + CHR$(&HE6) + CHR$(&H61)
   Zkbh$(2) = Zkbh$(2) + CHR$(&H24) + CHR$(&H7F)
   Zkbh$(2) = Zkbh$(2) + CHR$(&HE6) + CHR$(&H61)
   Zkbh$(2) = Zkbh$(2) + CHR$(&HB0) + CHR$(&H20)
   Zkbh$(2) = Zkbh$(2) + CHR$(&HE6) + CHR$(&H20)
   Zkbh$(2) = Zkbh$(2) + CHR$(&H1F)
   Zkbh$(2) = Zkbh$(2) + CHR$(&H5B)
   Zkbh$(2) = Zkbh$(2) + CHR$(&H58)
   Zkbh$(2) = Zkbh$(2) + CHR$(&HCF)

   DEF SEG = VARSEG(Zkbh$(0))
   CALL Absolute(seg1%, off1%, SADD(Zkbh$(0)))
   DEF SEG = VARSEG(Zkbh$(1))
   CALL Absolute(BYVAL VARSEG(Zkbh$(2)), BYVAL SADD(Zkbh$(2)), SADD(Zkbh$(1)))
   DEF SEG
   OUT &H21, (INP(&H21) AND (255 XOR 2))

   olkbseg% = seg1%
   olkboff% = off1%

   kbonflag% = -1

END SUB

SUB Wireframe

   FOR i% = 0 TO (nooftris3% - 1)

      ax = triangle3(i%).x1
      ay = triangle3(i%).y1
      bx = triangle3(i%).x2
      by = triangle3(i%).y2
      cx = triangle3(i%).x3
      cy = triangle3(i%).y3

      LINE (ax, ay)-(bx, by), 15
      LINE (bx, by)-(cx, cy), 15
      LINE (cx, cy)-(ax, ay), 15

   NEXT

END SUB

